var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
function tr(text) { return text; }
loader.register_task(loader.Stage.JAVASCRIPT_INITIALIZING, {
    name: "certificate accept tester",
    function: () => __awaiter(this, void 0, void 0, function* () {
        const certificate_accept = settings.static_global(Settings.KEY_CERTIFICATE_CALLBACK, undefined);
        const container_success = $("#container-success").hide();
        if (!certificate_accept) {
            loader.critical_error(tr("Missing certificate callback data"), tr("Please reconnect manually."));
            throw "missing data";
        }
        log.info(LogCategory.IPC, tr("Using this instance as certificate callback. ID: %s"), certificate_accept);
        try {
            yield bipc.get_handler().post_certificate_accpected(certificate_accept);
            log.info(LogCategory.IPC, tr("Other instance has acknowledged out work. Closing this window."));
            let seconds = 5;
            let interval_id;
            interval_id = setInterval(() => {
                seconds--;
                $("#time-left").text(seconds.toString());
                if (seconds <= 0) {
                    clearTimeout(interval_id);
                    log.info(LogCategory.GENERAL, tr("Closing window"));
                    window.close();
                    return;
                }
            }, 1000);
            container_success.show();
        }
        catch (error) {
            log.warn(LogCategory.IPC, tr("Failed to successfully post certificate accept status: %o"), error);
            loader.critical_error(tr("Failed to emit success!"), tr("Please reconnect manually."));
        }
    }),
    priority: 10
});
//Used by CertAccept popup
var bipc;
(function (bipc) {
    function uuidv4() {
        return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function (c) {
            const r = Math.random() * 16 | 0, v = c == 'x' ? r : (r & 0x3 | 0x8);
            return v.toString(16);
        });
    }
    class BasicIPCHandler {
        constructor() {
            this._channels = [];
            this._query_results = {};
            this._cert_accept_callbacks = {};
            this._cert_accept_succeeded = {};
        }
        setup() {
            this.unique_id = uuidv4(); /* lets get an unique identifier */
        }
        get_local_address() { return this.unique_id; }
        handle_message(message) {
            log.trace(LogCategory.IPC, tr("Received message %o"), message);
            if (message.receiver === BasicIPCHandler.BROADCAST_UNIQUE_ID) {
                if (message.type == "process-query") {
                    log.debug(LogCategory.IPC, tr("Received a device query from %s."), message.sender);
                    this.send_message("process-query-response", {
                        request_query_id: message.data.query_id,
                        request_timestamp: message.data.timestamp,
                        device_id: this.unique_id,
                        protocol: BasicIPCHandler.PROTOCOL_VERSION
                    }, message.sender);
                    return;
                }
            }
            else if (message.receiver === this.unique_id) {
                if (message.type == "process-query-response") {
                    const response = message.data;
                    if (this._query_results[response.request_query_id])
                        this._query_results[response.request_query_id].push(response);
                    else {
                        log.warn(LogCategory.IPC, tr("Received a query response for an unknown request."));
                    }
                    return;
                }
                else if (message.type == "certificate-accept-callback") {
                    const data = message.data;
                    if (!this._cert_accept_callbacks[data.request_id]) {
                        log.warn(LogCategory.IPC, tr("Received certificate accept callback for an unknown request ID."));
                        return;
                    }
                    this._cert_accept_callbacks[data.request_id]();
                    delete this._cert_accept_callbacks[data.request_id];
                    this.send_message("certificate-accept-succeeded", {}, message.sender);
                    return;
                }
                else if (message.type == "certificate-accept-succeeded") {
                    if (!this._cert_accept_succeeded[message.sender]) {
                        log.warn(LogCategory.IPC, tr("Received certificate accept succeeded, but haven't a callback."));
                        return;
                    }
                    this._cert_accept_succeeded[message.sender]();
                    return;
                }
                else if (message.type === "channel") {
                    const data = message.data;
                    let channel_invoked = false;
                    for (const channel of this._channels)
                        if (channel.channel_id === data.channel_id && (typeof (channel.target_id) === "undefined" || channel.target_id === message.sender)) {
                            if (channel.message_handler)
                                channel.message_handler(message.sender, data);
                            channel_invoked = true;
                        }
                    if (!channel_invoked) {
                        console.warn(tr("Received channel message for unknown channel (%s)"), data.channel_id);
                    }
                }
            }
        }
        create_channel(target_id, channel_id) {
            let channel = {
                target_id: target_id,
                channel_id: channel_id || uuidv4(),
                message_handler: undefined,
                send_message: (key, message) => {
                    if (!channel.target_id)
                        throw "channel has no target!";
                    this.send_message("channel", {
                        key: key,
                        message: message,
                        channel_id: channel.channel_id
                    }, channel.target_id);
                }
            };
            this._channels.push(channel);
            return channel;
        }
        channels() { return this._channels; }
        delete_channel(channel) {
            this._channels = this._channels.filter(e => e !== channel);
        }
        query_processes(timeout) {
            return __awaiter(this, void 0, void 0, function* () {
                const query_id = uuidv4();
                this._query_results[query_id] = [];
                this.send_message("process-query", {
                    query_id: query_id,
                    timestamp: Date.now()
                });
                yield new Promise(resolve => setTimeout(resolve, timeout || 250));
                const result = this._query_results[query_id];
                delete this._query_results[query_id];
                return result;
            });
        }
        register_certificate_accept_callback(callback) {
            const id = uuidv4();
            this._cert_accept_callbacks[id] = callback;
            return this.unique_id + ":" + id;
        }
        post_certificate_accpected(id, timeout) {
            return new Promise((resolve, reject) => {
                const data = id.split(":");
                const timeout_id = setTimeout(() => {
                    delete this._cert_accept_succeeded[data[0]];
                    clearTimeout(timeout_id);
                    reject("timeout");
                }, timeout || 250);
                this._cert_accept_succeeded[data[0]] = () => {
                    delete this._cert_accept_succeeded[data[0]];
                    clearTimeout(timeout_id);
                    resolve();
                };
                this.send_message("certificate-accept-callback", {
                    request_id: data[1]
                }, data[0]);
            });
        }
    }
    BasicIPCHandler.BROADCAST_UNIQUE_ID = "00000000-0000-4000-0000-000000000000";
    BasicIPCHandler.PROTOCOL_VERSION = 1;
    bipc.BasicIPCHandler = BasicIPCHandler;
    class BroadcastChannelIPC extends BasicIPCHandler {
        constructor() {
            super();
        }
        setup() {
            super.setup();
            this.channel = new BroadcastChannel(BroadcastChannelIPC.CHANNEL_NAME);
            this.channel.onmessage = this.on_message.bind(this);
            this.channel.onmessageerror = this.on_error.bind(this);
        }
        on_message(event) {
            if (typeof (event.data) !== "string") {
                log.warn(LogCategory.IPC, tr("Received message with an invalid type (%s): %o"), typeof (event.data), event.data);
                return;
            }
            let message;
            try {
                message = JSON.parse(event.data);
            }
            catch (error) {
                log.error(LogCategory.IPC, tr("Received an invalid encoded message: %o"), event.data);
                return;
            }
            super.handle_message(message);
        }
        on_error(event) {
            log.warn(LogCategory.IPC, tr("Received error: %o"), event);
        }
        send_message(type, data, target) {
            const message = {};
            message.sender = this.unique_id;
            message.receiver = target ? target : BasicIPCHandler.BROADCAST_UNIQUE_ID;
            message.timestamp = Date.now();
            message.type = type;
            message.data = data;
            this.channel.postMessage(JSON.stringify(message));
        }
    }
    BroadcastChannelIPC.CHANNEL_NAME = "TeaSpeak-Web";
    class MethodProxy {
        constructor(ipc_handler, connect_params) {
            this._proxied_methods = {};
            this._proxied_callbacks = {};
            this.ipc_handler = ipc_handler;
            this._ipc_parameters = connect_params;
            this._connected = false;
            this._slave = typeof (connect_params) !== "undefined";
            this._local = typeof (connect_params) !== "undefined" && connect_params.channel_id === "local" && connect_params.client_id === "local";
        }
        setup() {
            if (this._local) {
                this._connected = true;
                this.on_connected();
            }
            else {
                if (this._slave)
                    this._ipc_channel = this.ipc_handler.create_channel(this._ipc_parameters.client_id, this._ipc_parameters.channel_id);
                else
                    this._ipc_channel = this.ipc_handler.create_channel();
                this._ipc_channel.message_handler = this._handle_message.bind(this);
                if (this._slave)
                    this._ipc_channel.send_message("initialize", {});
            }
        }
        finalize() {
            if (!this._local) {
                if (this._connected)
                    this._ipc_channel.send_message("finalize", {});
                this.ipc_handler.delete_channel(this._ipc_channel);
                this._ipc_channel = undefined;
            }
            for (const promise of Object.values(this._proxied_callbacks))
                promise.reject("disconnected");
            this._proxied_callbacks = {};
            this._connected = false;
            this.on_disconnected();
        }
        register_method(method) {
            let method_name;
            if (typeof method === "function") {
                log.debug(LogCategory.IPC, tr("Registering method proxy for %s"), method.name);
                method_name = method.name;
            }
            else {
                log.debug(LogCategory.IPC, tr("Registering method proxy for %s"), method);
                method_name = method;
            }
            if (!this[method_name])
                throw "method is missing in current object";
            this._proxied_methods[method_name] = this[method_name];
            if (!this._local) {
                this[method_name] = (...args) => {
                    if (!this._connected)
                        return Promise.reject("not connected");
                    const proxy_callback = {
                        promise_id: uuidv4()
                    };
                    this._proxied_callbacks[proxy_callback.promise_id] = proxy_callback;
                    proxy_callback.promise = new Promise((resolve, reject) => {
                        proxy_callback.resolve = resolve;
                        proxy_callback.reject = reject;
                    });
                    this._ipc_channel.send_message("invoke", {
                        promise_id: proxy_callback.promise_id,
                        arguments: [...args],
                        method_name: method_name
                    });
                    return proxy_callback.promise;
                };
            }
        }
        _handle_message(remote_id, message) {
            if (message.key === "finalize") {
                this._handle_finalize();
            }
            else if (message.key === "initialize") {
                this._handle_remote_callback(remote_id);
            }
            else if (message.key === "invoke") {
                this._handle_invoke(message.message);
            }
            else if (message.key === "result") {
                this._handle_result(message.message);
            }
        }
        _handle_finalize() {
            this.on_disconnected();
            this.finalize();
            this._connected = false;
        }
        _handle_remote_callback(remote_id) {
            if (!this._ipc_channel.target_id) {
                if (this._slave)
                    throw "initialize wrong state!";
                this._ipc_channel.target_id = remote_id; /* now we're able to send messages */
                this.on_connected();
                this._ipc_channel.send_message("initialize", true);
            }
            else {
                if (!this._slave)
                    throw "initialize wrong state!";
                this.on_connected();
            }
            this._connected = true;
        }
        _send_result(promise_id, success, message) {
            this._ipc_channel.send_message("result", {
                promise_id: promise_id,
                result: message,
                success: success
            });
        }
        _handle_invoke(data) {
            if (this._proxied_methods[data.method_name])
                throw "we could not invoke a local proxied method!";
            if (!this[data.method_name]) {
                this._send_result(data.promise_id, false, "missing method");
                return;
            }
            try {
                log.info(LogCategory.IPC, tr("Invoking method %s with arguments: %o"), data.method_name, data.arguments);
                const promise = this[data.method_name](...data.arguments);
                promise.then(result => {
                    log.info(LogCategory.IPC, tr("Result: %o"), result);
                    this._send_result(data.promise_id, true, result);
                }).catch(error => {
                    this._send_result(data.promise_id, false, error);
                });
            }
            catch (error) {
                this._send_result(data.promise_id, false, error);
                return;
            }
        }
        _handle_result(data) {
            if (!this._proxied_callbacks[data.promise_id]) {
                console.warn(tr("Received proxy method result for unknown promise"));
                return;
            }
            const callback = this._proxied_callbacks[data.promise_id];
            delete this._proxied_callbacks[data.promise_id];
            if (data.success)
                callback.resolve(data.result);
            else
                callback.reject(data.result);
        }
        generate_connect_parameters() {
            if (this._slave)
                throw "only masters can generate connect parameters!";
            if (!this._ipc_channel)
                throw "please call setup() before";
            return {
                channel_id: this._ipc_channel.channel_id,
                client_id: this.ipc_handler.get_local_address()
            };
        }
        is_slave() { return this._local || this._slave; } /* the popout modal */
        is_master() { return this._local || !this._slave; } /* the host (teaweb application) */
    }
    bipc.MethodProxy = MethodProxy;
    let handler;
    function setup() {
        if (!supported())
            return;
        /* TODO: test for support */
        handler = new BroadcastChannelIPC();
        handler.setup();
    }
    bipc.setup = setup;
    function get_handler() {
        return handler;
    }
    bipc.get_handler = get_handler;
    function supported() {
        /* ios does not support this */
        return typeof (window.BroadcastChannel) !== "undefined";
    }
    bipc.supported = supported;
})(bipc || (bipc = {}));
//Used by CertAccept popup
if (typeof (customElements) !== "undefined") {
    try {
        class X_Properties extends HTMLElement {
        }
        class X_Property extends HTMLElement {
        }
        customElements.define('x-properties', X_Properties, { extends: 'div' });
        customElements.define('x-property', X_Property, { extends: 'div' });
    }
    catch (error) {
        console.warn("failed to define costum elements");
    }
}
class SettingsBase {
    static transformStO(input, _default, default_type) {
        default_type = default_type || typeof _default;
        if (typeof input === "undefined")
            return _default;
        if (default_type === "string")
            return input;
        else if (default_type === "number")
            return parseInt(input);
        else if (default_type === "boolean")
            return (input == "1" || input == "true");
        else if (default_type === "undefined")
            return input;
        return JSON.parse(input);
    }
    static transformOtS(input) {
        if (typeof input === "string")
            return input;
        else if (typeof input === "number")
            return input.toString();
        else if (typeof input === "boolean")
            return input ? "1" : "0";
        else if (typeof input === "undefined")
            return undefined;
        return JSON.stringify(input);
    }
    static resolveKey(key, _default, resolver, default_type) {
        let value = resolver(key.key);
        if (!value) {
            /* trying fallbacks */
            for (const fallback of key.fallback_keys || []) {
                value = resolver(fallback);
                if (typeof (value) === "string") {
                    /* fallback key succeeded */
                    const importer = (key.fallback_imports || {})[fallback];
                    if (importer)
                        return importer(value);
                    break;
                }
            }
        }
        if (typeof (value) !== 'string')
            return _default;
        return SettingsBase.transformStO(value, _default, default_type);
    }
    static keyify(key) {
        if (typeof (key) === "string")
            return { key: key };
        if (typeof (key) === "object" && key.key)
            return key;
        throw "key is not a key";
    }
}
SettingsBase.UPDATE_DIRECT = true;
class StaticSettings extends SettingsBase {
    constructor(_reserved = undefined) {
        super();
        if (_reserved && !StaticSettings._instance) {
            this._staticPropsTag = $("#properties");
            this.initializeStatic();
        }
        else {
            this._handle = StaticSettings.instance;
        }
    }
    static get instance() {
        if (!this._instance)
            this._instance = new StaticSettings(true);
        return this._instance;
    }
    initializeStatic() {
        let search;
        if (window.opener && window.opener !== window) {
            search = new URL(window.location.href).search;
        }
        else {
            search = location.search;
        }
        search.substr(1).split("&").forEach(part => {
            let item = part.split("=");
            $("<x-property></x-property>")
                .attr("key", item[0])
                .attr("value", item[1])
                .appendTo(this._staticPropsTag);
        });
    }
    static(key, _default, default_type) {
        if (this._handle)
            return this._handle.static(key, _default, default_type);
        key = StaticSettings.keyify(key);
        return StaticSettings.resolveKey(key, _default, key => {
            let result = this._staticPropsTag.find("[key='" + key + "']");
            if (result.length > 0)
                return decodeURIComponent(result.last().attr('value'));
            return false;
        }, default_type);
    }
    deleteStatic(key) {
        if (this._handle) {
            this._handle.deleteStatic(key);
            return;
        }
        key = StaticSettings.keyify(key);
        let result = this._staticPropsTag.find("[key='" + key.key + "']");
        if (result.length != 0)
            result.detach();
    }
}
class Settings extends StaticSettings {
    constructor() {
        super();
        this.cacheGlobal = {};
        this.updated = false;
        this.cacheGlobal = JSON.parse(localStorage.getItem("settings.global"));
        if (!this.cacheGlobal)
            this.cacheGlobal = {};
        this.saveWorker = setInterval(() => {
            if (this.updated)
                this.save();
        }, 5 * 1000);
    }
    static initialize() {
        settings = new Settings();
    }
    static_global(key, _default) {
        const actual_default = typeof (_default) === "undefined" && typeof (key) === "object" && 'default_value' in key ? key.default_value : _default;
        const default_object = { seed: Math.random() };
        let _static = this.static(key, default_object, typeof _default);
        if (_static !== default_object)
            return StaticSettings.transformStO(_static, actual_default);
        return this.global(key, actual_default);
    }
    global(key, _default) {
        const actual_default = typeof (_default) === "undefined" && typeof (key) === "object" && 'default_value' in key ? key.default_value : _default;
        return StaticSettings.resolveKey(Settings.keyify(key), actual_default, key => this.cacheGlobal[key]);
    }
    changeGlobal(key, value) {
        key = Settings.keyify(key);
        if (this.cacheGlobal[key.key] == value)
            return;
        this.updated = true;
        this.cacheGlobal[key.key] = StaticSettings.transformOtS(value);
        if (Settings.UPDATE_DIRECT)
            this.save();
    }
    save() {
        this.updated = false;
        let global = JSON.stringify(this.cacheGlobal);
        localStorage.setItem("settings.global", global);
        if (localStorage.save)
            localStorage.save();
    }
}
Settings.KEY_DISABLE_COSMETIC_SLOWDOWN = {
    key: 'disable_cosmetic_slowdown',
    description: 'Disable the cosmetic slowdows in some processes, like icon upload.'
};
Settings.KEY_DISABLE_CONTEXT_MENU = {
    key: 'disableContextMenu',
    description: 'Disable the context menu for the channel tree which allows to debug the DOM easier'
};
Settings.KEY_DISABLE_GLOBAL_CONTEXT_MENU = {
    key: 'disableGlobalContextMenu',
    description: 'Disable the general context menu prevention',
    default_value: false
};
Settings.KEY_DISABLE_UNLOAD_DIALOG = {
    key: 'disableUnloadDialog',
    description: 'Disables the unload popup on side closing'
};
Settings.KEY_DISABLE_VOICE = {
    key: 'disableVoice',
    description: 'Disables the voice bridge. If disabled, the audio and codec workers aren\'t required anymore'
};
Settings.KEY_DISABLE_MULTI_SESSION = {
    key: 'disableMultiSession',
    default_value: false,
    require_restart: true
};
Settings.KEY_LOAD_DUMMY_ERROR = {
    key: 'dummy_load_error',
    description: 'Triggers a loading error at the end of the loading process.'
};
/* Control bar */
Settings.KEY_CONTROL_MUTE_INPUT = {
    key: 'mute_input'
};
Settings.KEY_CONTROL_MUTE_OUTPUT = {
    key: 'mute_output'
};
Settings.KEY_CONTROL_SHOW_QUERIES = {
    key: 'show_server_queries'
};
Settings.KEY_CONTROL_CHANNEL_SUBSCRIBE_ALL = {
    key: 'channel_subscribe_all'
};
/* Connect parameters */
Settings.KEY_FLAG_CONNECT_DEFAULT = {
    key: 'connect_default'
};
Settings.KEY_CONNECT_ADDRESS = {
    key: 'connect_address'
};
Settings.KEY_CONNECT_PROFILE = {
    key: 'connect_profile'
};
Settings.KEY_CONNECT_USERNAME = {
    key: 'connect_username'
};
Settings.KEY_CONNECT_PASSWORD = {
    key: 'connect_password'
};
Settings.KEY_FLAG_CONNECT_PASSWORD = {
    key: 'connect_password_hashed'
};
Settings.KEY_CONNECT_HISTORY = {
    key: 'connect_history'
};
Settings.KEY_CERTIFICATE_CALLBACK = {
    key: 'certificate_callback'
};
/* sounds */
Settings.KEY_SOUND_MASTER = {
    key: 'audio_master_volume',
    default_value: 100
};
Settings.KEY_SOUND_MASTER_SOUNDS = {
    key: 'audio_master_volume_sounds',
    default_value: 100
};
Settings.KEY_CHAT_FIXED_TIMESTAMPS = {
    key: 'chat_fixed_timestamps',
    default_value: false,
    description: 'Enables fixed timestamps for chat messages and disabled the updating once (2 seconds ago... etc)'
};
Settings.KEY_CHAT_COLLOQUIAL_TIMESTAMPS = {
    key: 'chat_colloquial_timestamps',
    default_value: true,
    description: 'Enabled colloquial timestamp formatting like "Yesterday at ..." or "Today at ..."'
};
Settings.KEY_CHAT_COLORED_EMOJIES = {
    key: 'chat_colored_emojies',
    default_value: true,
    description: 'Enables colored emojies powered by Twemoji'
};
Settings.KEY_CHAT_TAG_URLS = {
    key: 'chat_tag_urls',
    default_value: true,
    description: 'Automatically link urls with [url]'
};
Settings.KEY_CHAT_ENABLE_MARKDOWN = {
    key: 'chat_enable_markdown',
    default_value: true,
    description: 'Enabled markdown chat support.'
};
Settings.KEY_CHAT_ENABLE_BBCODE = {
    key: 'chat_enable_bbcode',
    default_value: true,
    description: 'Enabled bbcode support in chat.'
};
Settings.KEY_SWITCH_INSTANT_CHAT = {
    key: 'switch_instant_chat',
    default_value: true,
    description: 'Directly switch to channel chat on channel select'
};
Settings.KEY_SWITCH_INSTANT_CLIENT = {
    key: 'switch_instant_client',
    default_value: true,
    description: 'Directly switch to client info on client select'
};
Settings.KEY_HOSTBANNER_BACKGROUND = {
    key: 'hostbanner_background',
    default_value: false,
    description: 'Enables a default background begind the hostbanner'
};
Settings.KEY_CHANNEL_EDIT_ADVANCED = {
    key: 'channel_edit_advanced',
    default_value: false,
    description: 'Edit channels in advanced mode with a lot more settings'
};
Settings.KEY_TEAFORO_URL = {
    key: "teaforo_url",
    default_value: "https://forum.teaspeak.de/"
};
Settings.KEY_FONT_SIZE = {
    key: "font_size"
};
Settings.KEY_ICON_SIZE = {
    key: "icon_size",
    default_value: 100
};
Settings.KEY_LAST_INVITE_LINK_TYPE = {
    key: "last_invite_link_type",
    default_value: "tea-web"
};
Settings.FN_INVITE_LINK_SETTING = name => {
    return {
        key: 'invite_link_setting_' + name
    };
};
Settings.FN_SERVER_CHANNEL_SUBSCRIBE_MODE = channel => {
    return {
        key: 'channel_subscribe_mode_' + channel
    };
};
Settings.FN_PROFILE_RECORD = name => {
    return {
        key: 'profile_record' + name
    };
};
Settings.KEYS = (() => {
    const result = [];
    for (const key in Settings) {
        if (!key.toUpperCase().startsWith("KEY_"))
            continue;
        if (key.toUpperCase() == "KEYS")
            continue;
        result.push(key);
    }
    return result;
})();
class ServerSettings extends SettingsBase {
    constructor() {
        super();
        this.cacheServer = {};
        this._server_settings_updated = false;
        this._destroyed = false;
        this._server_save_worker = setInterval(() => {
            if (this._server_settings_updated)
                this.save();
        }, 5 * 1000);
    }
    destroy() {
        this._destroyed = true;
        this._server_unique_id = undefined;
        this.cacheServer = undefined;
        clearInterval(this._server_save_worker);
        this._server_save_worker = undefined;
    }
    server(key, _default) {
        if (this._destroyed)
            throw "destroyed";
        return StaticSettings.resolveKey(Settings.keyify(key), _default, key => this.cacheServer[key]);
    }
    changeServer(key, value) {
        if (this._destroyed)
            throw "destroyed";
        key = Settings.keyify(key);
        if (this.cacheServer[key.key] == value)
            return;
        this._server_settings_updated = true;
        this.cacheServer[key.key] = StaticSettings.transformOtS(value);
        if (Settings.UPDATE_DIRECT)
            this.save();
    }
    setServer(server_unique_id) {
        if (this._destroyed)
            throw "destroyed";
        if (this._server_unique_id) {
            this.save();
            this.cacheServer = {};
            this._server_unique_id = undefined;
        }
        this._server_unique_id = server_unique_id;
        if (this._server_unique_id) {
            this.cacheServer = JSON.parse(localStorage.getItem("settings.server_" + server_unique_id));
            if (!this.cacheServer)
                this.cacheServer = {};
        }
    }
    save() {
        if (this._destroyed)
            throw "destroyed";
        this._server_settings_updated = false;
        if (this._server_unique_id) {
            let server = JSON.stringify(this.cacheServer);
            localStorage.setItem("settings.server_" + this._server_unique_id, server);
            if (localStorage.save)
                localStorage.save();
        }
    }
}
let settings;
//Used by CertAccept popup
if (!JSON.map_to) {
    JSON.map_to = function (object, json, variables, validator, variable_direction) {
        if (!validator)
            validator = (a, b) => true;
        if (!variables) {
            variables = [];
            if (!variable_direction || variable_direction == 0) {
                for (let field in json)
                    variables.push(field);
            }
            else if (variable_direction == 1) {
                for (let field in object)
                    variables.push(field);
            }
        }
        else if (!Array.isArray(variables)) {
            variables = [variables];
        }
        let updates = 0;
        for (let field of variables) {
            if (!json[field]) {
                console.trace(tr("Json does not contains %s"), field);
                continue;
            }
            if (!validator(field, json[field])) {
                console.trace(tr("Validator results in false for %s"), field);
                continue;
            }
            if (JSON.map_field_to(object, json[field], field))
                updates++;
        }
        return updates;
    };
}
if (!JSON.map_field_to) {
    JSON.map_field_to = function (object, value, field) {
        let field_type = typeof (object[field]);
        let new_object;
        if (field_type == "string" || field_type == "object" || field_type == "undefined")
            new_object = value;
        else if (field_type == "number")
            new_object = parseFloat(value);
        else if (field_type == "boolean")
            new_object = value == "1" || value == "true";
        else {
            console.warn(tr("Invalid object type %s for entry %s"), field_type, field);
            return false;
        }
        if (new_object === object[field])
            return false;
        object[field] = new_object;
        return true;
    };
}
if (!Array.prototype.remove) {
    Array.prototype.remove = function (elem) {
        const index = this.indexOf(elem, 0);
        if (index > -1) {
            this.splice(index, 1);
            return true;
        }
        return false;
    };
}
if (!Array.prototype.pop_front) {
    Array.prototype.pop_front = function () {
        if (this.length == 0)
            return undefined;
        return this.splice(0, 1)[0];
    };
}
if (!Array.prototype.last) {
    Array.prototype.last = function () {
        if (this.length == 0)
            return undefined;
        return this[this.length - 1];
    };
}
if (typeof ($) !== "undefined") {
    if (!$.spawn) {
        $.spawn = function (tagName) {
            return $(document.createElement(tagName));
        };
    }
    if (!$.fn.renderTag) {
        $.fn.renderTag = function (values) {
            let result;
            if (this.render) {
                result = $(this.render(values));
            }
            else {
                const template = window.jsrender.render[this.attr("id")];
                if (!template) {
                    console.error("Tried to render template %o, but template is not available!", this.attr("id"));
                    throw "missing template " + this.attr("id");
                }
                /*
                result = window.jsrender.templates("tmpl_permission_entry", $("#tmpl_permission_entry").html());
                result = window.jsrender.templates("xxx", this.html());
                */
                result = template(values);
                result = $(result);
            }
            result.find("node").each((index, element) => {
                $(element).replaceWith(values[$(element).attr("key")] || (values[0] || [])[$(element).attr("key")]);
            });
            return result;
        };
    }
    if (!$.fn.hasScrollBar)
        $.fn.hasScrollBar = function (direction) {
            if (this.length <= 0)
                return false;
            const scroll_height = this.get(0).scrollHeight > this.height();
            const scroll_width = this.get(0).scrollWidth > this.width();
            if (typeof (direction) === "string") {
                if (direction === "height")
                    return scroll_height;
                if (direction === "width")
                    return scroll_width;
            }
            return scroll_width || scroll_height;
        };
    if (!$.fn.visible_height)
        $.fn.visible_height = function () {
            const original_style = this.attr("style");
            this.css({
                position: 'absolute!important',
                visibility: 'hidden!important',
                display: 'block!important'
            });
            const result = this.height();
            this.attr("style", original_style || "");
            return result;
        };
    if (!$.fn.visible_width)
        $.fn.visible_width = function () {
            const original_style = this.attr("style");
            this.css({
                position: 'absolute!important',
                visibility: 'hidden!important',
                display: 'block!important'
            });
            const result = this.width();
            this.attr("style", original_style || "");
            return result;
        };
    if (!$.fn.firstParent)
        $.fn.firstParent = function (selector) {
            if (this.is(selector))
                return this;
            return this.parent(selector);
        };
}
if (!String.prototype.format) {
    String.prototype.format = function () {
        const args = arguments;
        let array = args.length == 1 && $.isArray(args[0]);
        return this.replace(/\{\{|\}\}|\{(\d+)\}/g, function (m, n) {
            if (m == "{{") {
                return "{";
            }
            if (m == "}}") {
                return "}";
            }
            return array ? args[0][n] : args[n];
        });
    };
}
function concatenate(resultConstructor, ...arrays) {
    let totalLength = 0;
    for (const arr of arrays) {
        totalLength += arr.length;
    }
    const result = new resultConstructor(totalLength);
    let offset = 0;
    for (const arr of arrays) {
        result.set(arr, offset);
        offset += arr.length;
    }
    return result;
}
function formatDate(secs) {
    let years = Math.floor(secs / (60 * 60 * 24 * 365));
    let days = Math.floor(secs / (60 * 60 * 24)) % 365;
    let hours = Math.floor(secs / (60 * 60)) % 24;
    let minutes = Math.floor(secs / 60) % 60;
    let seconds = Math.floor(secs % 60);
    let result = "";
    if (years > 0)
        result += years + " " + tr("years") + " ";
    if (years > 0 || days > 0)
        result += days + " " + tr("days") + " ";
    if (years > 0 || days > 0 || hours > 0)
        result += hours + " " + tr("hours") + " ";
    if (years > 0 || days > 0 || hours > 0 || minutes > 0)
        result += minutes + " " + tr("minutes") + " ";
    if (years > 0 || days > 0 || hours > 0 || minutes > 0 || seconds > 0)
        result += seconds + " " + tr("seconds") + " ";
    else
        result = tr("now") + " ";
    return result.substr(0, result.length - 1);
}
function calculate_width(text) {
    let element = $.spawn("div");
    element.text(text)
        .css("display", "none")
        .css("margin", 0);
    $("body").append(element);
    let size = element.width();
    element.detach();
    return size;
}
//Used by CertAccept popup
var LogCategory;
(function (LogCategory) {
    LogCategory[LogCategory["CHANNEL"] = 0] = "CHANNEL";
    LogCategory[LogCategory["CHANNEL_PROPERTIES"] = 1] = "CHANNEL_PROPERTIES";
    LogCategory[LogCategory["CLIENT"] = 2] = "CLIENT";
    LogCategory[LogCategory["BOOKMARKS"] = 3] = "BOOKMARKS";
    LogCategory[LogCategory["SERVER"] = 4] = "SERVER";
    LogCategory[LogCategory["PERMISSIONS"] = 5] = "PERMISSIONS";
    LogCategory[LogCategory["GENERAL"] = 6] = "GENERAL";
    LogCategory[LogCategory["NETWORKING"] = 7] = "NETWORKING";
    LogCategory[LogCategory["VOICE"] = 8] = "VOICE";
    LogCategory[LogCategory["CHAT"] = 9] = "CHAT";
    LogCategory[LogCategory["AUDIO"] = 10] = "AUDIO";
    LogCategory[LogCategory["I18N"] = 11] = "I18N";
    LogCategory[LogCategory["IPC"] = 12] = "IPC";
    LogCategory[LogCategory["IDENTITIES"] = 13] = "IDENTITIES";
    LogCategory[LogCategory["STATISTICS"] = 14] = "STATISTICS";
})(LogCategory || (LogCategory = {}));
var log;
(function (log_1) {
    let LogType;
    (function (LogType) {
        LogType[LogType["TRACE"] = 0] = "TRACE";
        LogType[LogType["DEBUG"] = 1] = "DEBUG";
        LogType[LogType["INFO"] = 2] = "INFO";
        LogType[LogType["WARNING"] = 3] = "WARNING";
        LogType[LogType["ERROR"] = 4] = "ERROR";
    })(LogType = log_1.LogType || (log_1.LogType = {}));
    let category_mapping = new Map([
        [LogCategory.CHANNEL, "Channel    "],
        [LogCategory.CHANNEL_PROPERTIES, "Channel    "],
        [LogCategory.CLIENT, "Client     "],
        [LogCategory.SERVER, "Server     "],
        [LogCategory.BOOKMARKS, "Bookmark   "],
        [LogCategory.PERMISSIONS, "Permission "],
        [LogCategory.GENERAL, "General    "],
        [LogCategory.NETWORKING, "Network    "],
        [LogCategory.VOICE, "Voice      "],
        [LogCategory.AUDIO, "Audio      "],
        [LogCategory.CHANNEL, "Chat       "],
        [LogCategory.I18N, "I18N       "],
        [LogCategory.IDENTITIES, "Identities "],
        [LogCategory.IPC, "IPC        "],
        [LogCategory.STATISTICS, "Statistics "]
    ]);
    log_1.enabled_mapping = new Map([
        [LogCategory.CHANNEL, true],
        [LogCategory.CHANNEL_PROPERTIES, false],
        [LogCategory.CLIENT, true],
        [LogCategory.SERVER, true],
        [LogCategory.BOOKMARKS, true],
        [LogCategory.PERMISSIONS, true],
        [LogCategory.GENERAL, true],
        [LogCategory.NETWORKING, true],
        [LogCategory.VOICE, true],
        [LogCategory.AUDIO, true],
        [LogCategory.CHAT, true],
        [LogCategory.I18N, false],
        [LogCategory.IDENTITIES, true],
        [LogCategory.IPC, true],
        [LogCategory.STATISTICS, true]
    ]);
    //Values will be overridden by initialize()
    log_1.level_mapping = new Map([
        [LogType.TRACE, true],
        [LogType.DEBUG, true],
        [LogType.INFO, true],
        [LogType.WARNING, true],
        [LogType.ERROR, true]
    ]);
    let GroupMode;
    (function (GroupMode) {
        GroupMode[GroupMode["NATIVE"] = 0] = "NATIVE";
        GroupMode[GroupMode["PREFIX"] = 1] = "PREFIX";
    })(GroupMode || (GroupMode = {}));
    const group_mode = GroupMode.PREFIX;
    //Category Example: <url>?log.i18n.enabled=0
    //Level Example A: <url>?log.level.trace.enabled=0
    //Level Example B: <url>?log.level=0
    function initialize(default_level) {
        for (const category of Object.keys(LogCategory).map(e => parseInt(e))) {
            if (isNaN(category))
                continue;
            const category_name = LogCategory[category].toLowerCase();
            log_1.enabled_mapping.set(category, settings.static_global("log." + category_name.toLowerCase() + ".enabled", log_1.enabled_mapping.get(category)));
        }
        const base_level = settings.static_global("log.level", default_level);
        for (const level of Object.keys(LogType).map(e => parseInt(e))) {
            if (isNaN(level))
                continue;
            const level_name = LogType[level].toLowerCase();
            log_1.level_mapping.set(level, settings.static_global("log." + level_name + ".enabled", level >= base_level));
        }
    }
    log_1.initialize = initialize;
    function logDirect(type, message, ...optionalParams) {
        if (!log_1.level_mapping.get(type))
            return;
        switch (type) {
            case LogType.TRACE:
            case LogType.DEBUG:
                console.debug(message, ...optionalParams);
                break;
            case LogType.INFO:
                console.log(message, ...optionalParams);
                break;
            case LogType.WARNING:
                console.warn(message, ...optionalParams);
                break;
            case LogType.ERROR:
                console.error(message, ...optionalParams);
                break;
        }
    }
    function log(type, category, message, ...optionalParams) {
        if (!log_1.enabled_mapping.get(category))
            return;
        optionalParams.unshift(category_mapping.get(category));
        message = "[%s] " + message;
        logDirect(type, message, ...optionalParams);
    }
    log_1.log = log;
    function trace(category, message, ...optionalParams) {
        log(LogType.TRACE, category, message, ...optionalParams);
    }
    log_1.trace = trace;
    function debug(category, message, ...optionalParams) {
        log(LogType.DEBUG, category, message, ...optionalParams);
    }
    log_1.debug = debug;
    function info(category, message, ...optionalParams) {
        log(LogType.INFO, category, message, ...optionalParams);
    }
    log_1.info = info;
    function warn(category, message, ...optionalParams) {
        log(LogType.WARNING, category, message, ...optionalParams);
    }
    log_1.warn = warn;
    function error(category, message, ...optionalParams) {
        log(LogType.ERROR, category, message, ...optionalParams);
    }
    log_1.error = error;
    function group(level, category, name, ...optionalParams) {
        name = "[%s] " + name;
        optionalParams.unshift(category_mapping.get(category));
        return new Group(group_mode, level, category, name, optionalParams);
    }
    log_1.group = group;
    function table(level, category, title, arguments) {
        if (group_mode == GroupMode.NATIVE) {
            console.groupCollapsed(title);
            console.table(arguments);
            console.groupEnd();
        }
        else {
            if (!log_1.enabled_mapping.get(category) || !log_1.level_mapping.get(level))
                return;
            logDirect(level, tr("Snipped table \"%s\""), title);
        }
    }
    log_1.table = table;
    class Group {
        constructor(mode, level, category, name, optionalParams, owner = undefined) {
            this.owner = undefined;
            this._collapsed = false;
            this.initialized = false;
            this.level = level;
            this.mode = mode;
            this.category = category;
            this.name = name;
            this.optionalParams = optionalParams;
            this.enabled = log_1.enabled_mapping.get(category);
        }
        group(level, name, ...optionalParams) {
            return new Group(this.mode, level, this.category, name, optionalParams, this);
        }
        collapsed(flag = true) {
            this._collapsed = flag;
            return this;
        }
        log(message, ...optionalParams) {
            if (!this.enabled)
                return this;
            if (!this.initialized) {
                if (this.mode == GroupMode.NATIVE) {
                    if (this._collapsed && console.groupCollapsed)
                        console.groupCollapsed(this.name, ...this.optionalParams);
                    else
                        console.group(this.name, ...this.optionalParams);
                }
                else {
                    this._log_prefix = "  ";
                    let parent = this.owner;
                    while (parent) {
                        if (parent.mode == GroupMode.PREFIX)
                            this._log_prefix = this._log_prefix + parent._log_prefix;
                        else
                            break;
                    }
                }
                this.initialized = true;
            }
            if (this.mode == GroupMode.NATIVE)
                logDirect(this.level, message, ...optionalParams);
            else {
                logDirect(this.level, "[%s] " + this._log_prefix + message, category_mapping.get(this.category), ...optionalParams);
            }
            return this;
        }
        end() {
            if (this.initialized) {
                if (this.mode == GroupMode.NATIVE)
                    console.groupEnd();
            }
        }
        get prefix() {
            return this._log_prefix;
        }
        set prefix(prefix) {
            this._log_prefix = prefix;
        }
    }
    log_1.Group = Group;
})(log || (log = {}));
var LogType = log.LogType;
